use std::{net::Ipv4Addr, path::Path};

use nom::{
    bytes::complete::take,
    number::complete::{be_u16, be_u32, be_u8},
    IResult,
};

#[derive(Debug)]
struct DhcpMessage<'a> {
    operation: Operation,
    hardware_type: HardwareType,
    hardware_address_length: u8,
    hops: u8,
    transaction_id: u32,
    seconds: u16,
    flags: u16,
    client_address: Ipv4Addr,
    server_address: Ipv4Addr,
    gateway_address: Ipv4Addr,
    client_hardware_address: &'a [u8],
    server_hostname: &'a str,
    boot_file_name: &'a Path,
}

#[derive(Debug)]
enum Operation {
    Request = 1,
    Reply = 2,
}

impl From<u8> for Operation {
    fn from(value: u8) -> Self {
        match value {
            1 => Operation::Request,
            2 => Operation::Reply,
            _ => todo!(),
        }
    }
}

#[derive(Debug)]
enum HardwareType {
    Reserved(u8),
    Unassigned(u8),
    Ethernet,
}

impl From<u8> for HardwareType {
    fn from(value: u8) -> Self {
        use HardwareType::*;
        match value {
            a @ 0 => Reserved(a),
            1 => Ethernet,
            v => Unassigned(v),
        }
    }
}

fn parse_dhcp_operation(i: &[u8]) -> IResult<&[u8], Operation> {
    be_u8(i).map(|(i, v)| (i, v.into()))
}

fn parse_dhcp_hardware_type(i: &[u8]) -> IResult<&[u8], HardwareType> {
    be_u8(i).map(|(i, v)| (i, v.into()))
}

fn parse_ipv4_address(i: &[u8]) -> IResult<&[u8], Ipv4Addr> {
    let (i, val) = take(4usize)(i)?;
    let ip = Ipv4Addr::new(val[0], val[1], val[2], val[3]);
    Ok((i, ip))
}

fn parse_dhcp_packet(packet: &[u8]) -> IResult<&[u8], DhcpMessage> {
    let (packet, operation) = parse_dhcp_operation(packet)?;
    let (packet, hardware_type) = parse_dhcp_hardware_type(packet)?;
    let (packet, hardware_address_length) = be_u8(packet)?;
    let (packet, hops) = be_u8(packet)?;
    let (packet, transaction_id) = be_u32(packet)?;
    let (packet, seconds) = be_u16(packet)?;
    let (packet, flags) = be_u16(packet)?;
    let (packet, client_address) = parse_ipv4_address(packet)?;
    let (packet, server_address) = parse_ipv4_address(packet)?;
    let (packet, gateway_address) = parse_ipv4_address(packet)?;
    let (packet, client_hardware_address) = take(6usize)(packet)?;

    Ok((
        packet,
        DhcpMessage {
            operation,
            hardware_type,
            hardware_address_length,
            hops,
            transaction_id,
            seconds,
            flags,
            client_address,
            server_address,
            gateway_address,
            client_hardware_address,
            server_hostname: "",
            boot_file_name: Path::new(""),
        },
    ))
}

#[cfg(test)]
mod tests {
    static DHCP_PACKET: &'static [u8] = include_bytes!("../dhcp_discover.bin");
    static DHCP_PACKET2: &'static [u8] = &[
        0x01, 0x01, 0x06, 0x05, 0x00, 0x00, 0x00, 0x01, 0x00, 0x02, 0x00, 0x01, 0x7f, 0x00, 0x00,
        0x01, 0x7f, 0x00, 0x00, 0x01, 0x7f, 0x00, 0x00, 0x01, 0x54, 0x05, 0xdb, 0xa6, 0x96, 0xd0,
    ];

    #[test]
    fn parses() {
        let res = super::parse_dhcp_packet(DHCP_PACKET2).unwrap();
        dbg!(res);
    }
}
